# coding: utf-8
#  Copyright (c) 2020 Kumagai group.
import argparse
import sys
from fractions import Fraction
from pathlib import Path

import crystal_toolkit.components as ctc
import numpy as np
from crystal_toolkit.components import StructureMoleculeComponent
from crystal_toolkit.helpers.layouts import *
from dash import Dash
from pymatgen.core import Structure, Lattice, Molecule
from pymatgen.ext.matproj import MPRester

from pymatgen.analysis.chemenv.coordination_environments.chemenv_strategies import \
    SimplestChemenvStrategy
from pymatgen.analysis.chemenv.coordination_environments.coordination_geometries import \
    AllCoordinationGeometries
from pymatgen.analysis.chemenv.coordination_environments.coordination_geometry_finder import \
    LocalGeometryFinder
from pymatgen.analysis.chemenv.coordination_environments.structure_environments import \
    LightStructureEnvironments
from pymatgen.analysis.graphs import MoleculeGraph
from pymatgen.analysis.local_env import CrystalNN
from pymatgen.analysis.structure_matcher import StructureMatcher
from vise.analyzer.dash_components.band_dos_dash import BandDosComponent
from vise.analyzer.dash_components.structure_component import StructureComponent
from vise.analyzer.dos_data import DosPlotData
from vise.analyzer.plot_band import BandPlotInfo
from vise.defaults import defaults
from vise.util.structure_symmetrizer import StructureSymmetrizer


def create_ctk(struct_component,
               crystal_symmetry,
               sites,
               band_dos_component):
    box_size = "30vmin"
    _layout = Container(
        [
            Section(
                [
                    Columns(
                        [
                            Column(
                                [struct_component.title_layout()]
                            )
                        ]
                    ),
                    Columns(
                        [
                            Column(
                                [
                                    Box(
                                        struct_component.layout(size="100%"),
                                        style={
                                            "width": box_size,
                                            "height": box_size,
                                            "minWidth": "300px",
                                            "minHeight": "300px",
                                            "maxWidth": "600px",
                                            "maxHeight": "600px",
                                            "overflow": "hidden",
                                            "padding": "0.25rem",
                                            "marginBottom": "0.5rem",
                                        },
                                    ),
                                    html.Div(
                                        [
                                            html.Div(
                                                struct_component.legend_layout(),
                                                style={"float": "left"},
                                            ),
                                        ],
                                        style={
                                            "width": box_size,
                                            "minWidth": "300px",
                                            "marginBottom": "40px",
                                        },
                                    ),
                                ],
                                narrow=True,
                            ),
                            Column(
                                [
                                    crystal_symmetry,
                                ],
                                style={"width": box_size, "max-width": box_size},
                            ),
                        ],
                        desktop_only=False,
                        centered=False,
                    ),
                    Columns(Column([sites])),
                ]
            ),
            Section([band_dos_component.layout()]
            ),
        ]
    )

    return _layout


def symmetry_layout(structure: Structure):
    data = dict()
    symmetrizer = StructureSymmetrizer(structure)
    data["Space Group"] = f"{symmetrizer.spglib_sym_data['international']} " \
                          f"({symmetrizer.sg_number})"
    data["Point Group"] = symmetrizer.point_group

    datalist = get_data_list(data)

    return Columns([Column([H4("Crystal symmetry"), datalist])])


def mpid_and_link(symmetrizer: StructureSymmetrizer):

    reduced_formula = symmetrizer.structure.composition.reduced_formula
    criteria = {"reduced_cell_formula": reduced_formula,
                "spacegroup.number": symmetrizer.sg_number}
    properties = ["task_id", "structure"]
    with MPRester() as m:
        for doc in m.query(criteria, properties):
            sm = StructureMatcher()
            if sm.fit(doc["structure"], symmetrizer.structure):
                mpid = doc["task_id"]
                break
        else:
            return H4("None")

    return dcc.Link(f'mp-id {mpid}',
                    href=f'https://materialsproject.org/materials/{mpid}/',
                    style={'font-weight': 'bold', "font-size": "20px"})


def pretty_frac_format(x):
    x = x % 1
    fraction = Fraction(x).limit_denominator(8)
    if np.allclose(x, 1):
        x_str = "0"
    elif not np.allclose(x, float(fraction)):
        x = np.around(x, decimals=3)
        x_str = f"{x:.3g}"
    else:
        x_str = str(fraction)
    return x_str


def site_layout(symmetrizer: StructureSymmetrizer):
    columns = []
    distance_cutoff = 2.0
    angle_cutoff = 0.3
    cnn = CrystalNN()
    nn_info = cnn.get_all_nn_info(structure=symmetrizer.structure)

    for name, site in symmetrizer.sites.items():
        wyckoff_contents = []
        data = dict()
        data["Site Symmetry"] = site.site_symmetry
        data["Wyckoff Letter"] = site.wyckoff_letter
        wyckoff_contents.append(html.Label(f"{name}", className="mpc-label"))
        site_data = []
        repr_idx = site.equivalent_atoms[0]
        for idx in site.equivalent_atoms:
            coords = symmetrizer.structure[idx].frac_coords
            site_data += [(pretty_frac_format(coords[0]),
                           pretty_frac_format(coords[1]),
                           pretty_frac_format(coords[2]))]

        # lgf = LocalGeometryFinder()
        # lgf.setup_structure(structure=symmetrizer.structure)

        # se = lgf.compute_structure_environments(
        #     maximum_distance_factor=distance_cutoff + 0.01)
        # strategy = SimplestChemenvStrategy(
        #     distance_cutoff=distance_cutoff, angle_cutoff=angle_cutoff
        # )
        # lse = LightStructureEnvironments.from_structure_environments(
        #     strategy=strategy, structure_environments=se)

    # represent the local environment as a molecule

        mol = Molecule.from_sites(
            [symmetrizer.structure[repr_idx]] + [i["site"] for i in nn_info[repr_idx]]
        )
        mol = mol.get_centered_molecule()
        mg = MoleculeGraph.with_empty_graph(molecule=mol)
        for i in range(1, len(mol)):
            mg.add_edge(0, i)

        view = html.Div(
            [
                StructureMoleculeComponent(
                    struct_or_mol=mg,
                    disable_callbacks=True,
                    id=f"{symmetrizer.structure.composition.reduced_formula}_site_{repr_idx}",
                    scene_settings={"enableZoom": False, "defaultZoom": 0.6},
                )._sub_layouts["struct"]
            ],
            style={"width": "300px", "height": "300px"},
        )

        # env = lse.coordination_environments[repr_idx]
        # all_ce = AllCoordinationGeometries()
        # co = all_ce.get_geometry_from_mp_symbol(env[0]["ce_symbol"])
        # name = co.name

        data.update(
            {
                # "Environment": name,
                # "IUPAC Symbol": co.IUPAC_symbol_str,
                "Structure": view,
            }
        )

        wyckoff_contents.append(get_data_list(data))
        wyckoff_contents.append(Reveal(get_table(site_data),
                                       title=f"Positions ({len(site_data)})",
                                       id=f"{name}-positions"))
        columns.append(Column(html.Div(wyckoff_contents)))
    _layout = Columns(columns)

    return Columns([Column([H4("Sites"), html.Div(_layout)])])


def create_app(structure: Structure, dos_plot_data, band_plot_data):
    app = Dash(suppress_callback_exceptions=True,
               assets_folder=SETTINGS.ASSETS_PATH)

    structure_component = StructureComponent(structure)
    comp = structure.composition.reduced_formula
    band_dos_component = BandDosComponent(dos_plot_data, band_plot_data, id=f"band_dos_{comp}",)
    symmetrizer = StructureSymmetrizer(structure)
    delayed_layout = create_ctk(structure_component,
                                symmetry_layout(structure),
                                site_layout(symmetrizer),
                                band_dos_component)
    ctc.register_crystal_toolkit(layout=delayed_layout, app=app, cache=None)

    return app


def parse_args(args):
    parser = argparse.ArgumentParser(description="")
    parser.add_argument("-b", "--band", type=Path)
    parser.add_argument("-d", "--dos", type=Path)
    parser.add_argument("-p", "--port", type=int)
    return parser.parse_args(args)


if __name__ == "__main__":
    args = parse_args(sys.argv[1:])
    structure = Structure.from_file("POSCAR")
    dos_plot_data = loadfn(args.dos / "dos_plot_data.json")
    band_plot_data = loadfn(args.band / "band_plot_info.json")
    my_app = create_app(structure, dos_plot_data, band_plot_data)
    my_app.run_server(debug=True, port=args.port)
    # symmetrizer = StructureSymmetrizer(structure)
    #
    # app = Dash(suppress_callback_exceptions=True,
    #            assets_folder=SETTINGS.ASSETS_PATH)
    # ctc.register_crystal_toolkit(layout=html.Div([mpid_and_link(symmetrizer)]),
    #                              app=app, cache=None)
    # app.run_server(debug=True, port=args.port)




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import re
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional

from monty.serialization import loadfn


class ToJsonFileMixIn(ABC):
    def to_json_file(self, filename: Optional[str] = None) -> None:
        filename = filename or self._json_filename
        Path(filename).write_text(self.to_json())

    @abstractmethod
    def to_json(self):
        pass

    @property
    def _json_filename(self):
        """ ClassForThis -> class_for_this.json
        https://stackoverflow.com/questions/7322028/how-to-replace-uppercase-with-underscore
        """
        class_name = self.__class__.__name__
        return re.sub("(?<!^)(?=[A-Z])", "_", class_name).lower() + ".json"


class ToYamlFileMixIn(ABC):

    def to_yaml_file(self, filename: Optional[str] = None) -> None:
        filename = filename or self._yaml_filename()
        Path(filename).write_text(self.to_yaml())

    @abstractmethod
    def to_yaml(self):
        pass

    @classmethod
    def from_yaml(cls, filename: str = None):
        d = loadfn(filename or cls._yaml_filename())
        return cls(**d)

    @classmethod
    def _yaml_filename(cls):
        """ ClassForThis -> class_for_this.json
        https://stackoverflow.com/questions/7322028/how-to-replace-uppercase-with-underscore
        """
        class_name = cls.__name__
        return re.sub("(?<!^)(?=[A-Z])", "_", class_name).lower() + ".yaml"





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import dataclasses
import json
import sys
from difflib import Differ
from pathlib import Path

from monty.json import MSONable, MontyDecoder
from monty.serialization import loadfn
import numpy as np
from py._path.local import LocalPath
from vise.util.mix_in import ToYamlFileMixIn


def assert_msonable(obj):
    assert isinstance(obj, MSONable)
    assert obj.as_dict() == obj.__class__.from_dict(obj.as_dict()).as_dict()
    assert json.loads(obj.to_json(), cls=MontyDecoder)


def assert_json_roundtrip(obj, tmpdir):
    tmpdir.chdir()
    obj.to_json_file("a.json")
    actual = loadfn("a.json").as_dict()
    expected = obj.as_dict()
    for k, v in actual.items():
        try:
            assert v == expected[k]
        except AssertionError:
            print(k)
            print(v)
            print(expected[k])


def assert_yaml_roundtrip(obj: ToYamlFileMixIn,
                          tmpdir: LocalPath,
                          expected_text: str,
                          compare_dict: bool = True,
                          compare_items: bool = True):
    tmpdir.chdir()
    obj.to_yaml_file("a.yaml")
    actual_text = Path("a.yaml").read_text()
    try:
        assert actual_text == expected_text
    except AssertionError:
        print(tmpdir)
        a = actual_text.split("\n")
        b = expected_text.split("\n")
        sys.stdout.writelines(list(Differ().compare(a, b)))
        raise

    if compare_dict:
        try:
            actual = obj.from_yaml("a.yaml").as_dict()
            expected = obj.as_dict()
        except AttributeError:
            from dataclasses import asdict
            actual = asdict(obj.from_yaml("a.yaml"))
            expected = asdict(obj)
        try:
            assert len(actual) == len(expected)
        except AssertionError:
            print(tmpdir)
            raise
    else:
        actual = obj.from_yaml("a.yaml")
        print("actual", actual)

    if compare_items is True:
        for k, v in actual.items():
            try:
                assert v == expected[k]
            except AssertionError:
                print(tmpdir)
                print(f"key: {k}, actual: {v}, expected: {expected[k]}")
                raise


def assert_dataclass_almost_equal(actual, expected, digit=8):
    assert actual.__class__ == expected.__class__

    for k, v1 in actual.__dict__.items():
        print(f"key {k}")
        v2 = expected.__getattribute__(k)
        assert_attribute_almost_same(v1, v2, digit)


def assert_attribute_almost_same(v1, v2, digit):
    if isinstance(v1, float):
        try:
            assert round(v1, digit) == round(v2, digit)
        except AssertionError:
            print(v1, v2)
            raise
    elif isinstance(v1, tuple) or isinstance(v1, list):
        for vv1, vv2 in zip(v1, v2):
            assert_attribute_almost_same(vv1, vv2, digit)
    elif dataclasses.is_dataclass(v1):
        assert_dataclass_almost_equal(v1, v2, digit)
    else:
        try:
            if isinstance(v1, np.ndarray):
                np.testing.assert_array_almost_equal(v1, v2)
            else:
                assert v1 == v2
        except AssertionError:
            print(v1, v2)
            raise



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from argparse import Namespace
from copy import deepcopy
from pathlib import Path

import yaml
from pymatgen.core import Structure
from pymatgen.ext.matproj import MPRester
from pymatgen.io.vasp import Vasprun, Outcar
from vise.analyzer.plot_absorption_coeff import AbsorptionCoeffMplPlotter
from vise.analyzer.plot_band import BandPlotter
from vise.analyzer.plot_dos import DosPlotter
from vise.analyzer.vasp.band_edge_properties import VaspBandEdgeProperties
from vise.analyzer.vasp.dos_data import DosDataFromVasp
from vise.analyzer.vasp.make_diele_func import make_diele_func
from vise.analyzer.vasp.make_effective_mass import make_effective_mass
from vise.analyzer.vasp.plot_band import BandPlotInfoFromVasp
from vise.atom_energies.make_atom_vasp_set import make_atom_poscar_dirs
from vise.cli.main_tools import potcar_str2dict, list2dict
from vise.defaults import defaults
from vise.input_set.datasets.dataset_util import all_incar_flags
from vise.input_set.input_options import CategorizedInputOptions, \
    assignable_option_set
from vise.input_set.kpoints_mode import KpointsMode
from vise.input_set.prior_info import prior_info_from_calc_dir, PriorInfo
from vise.input_set.task import Task
from vise.input_set.vasp_input_files import VaspInputFiles
from vise.util.file_transfer import FileTransfers
from vise.util.logger import get_logger
from vise.util.structure_symmetrizer import StructureSymmetrizer

logger = get_logger(__name__)


def structure_info(args: Namespace) -> None:
    input_structure = Structure.from_file(args.poscar)
    symmetrizer = StructureSymmetrizer(input_structure,
                                       symprec=args.symprec,
                                       angle_tolerance=args.angle_tolerance)
    if args.show_primitive:
        primitive = symmetrizer.primitive
        if primitive != input_structure:
            print(primitive.to(fmt="poscar"))
        else:
            logger.info("Input structure is a primitive cell.")
    elif args.show_conventional:
        conventional = symmetrizer.conventional
        if conventional != input_structure:
            print(conventional.to(fmt="poscar"))
        else:
            logger.info("Input structure is a conventional cell.")
    else:
        print(symmetrizer)


def get_poscar_from_mp(args: Namespace) -> None:
    s = MPRester().get_structure_by_material_id(args.mpid)
    s.to(fmt="poscar", filename=args.poscar)
    data = MPRester().get_data(args.mpid)[0]
    d = {"total_magnetization": data["total_magnetization"],
         "band_gap": data["band_gap"],
         "data_source": args.mpid,
         "icsd_ids": data["icsd_ids"]}
    args.prior_info.write_text(yaml.dump(d), None)


class VaspSet:
    def __init__(self, args: Namespace):
        self.args = args

        try:
            self._prior_info = PriorInfo.load_yaml()
        except FileNotFoundError:
            self._prior_info = PriorInfo()
        self.task = Task.cluster_opt if self._prior_info.is_cluster \
            else args.task

        options = CategorizedInputOptions(
            structure=self._structure(),
            task=self.task,
            xc=args.xc,
            kpt_density=args.kpt_density,
            overridden_potcar=self._overridden_potcar(),
            **self._option_kwargs())

        vif = VaspInputFiles(options, self._overridden_incar_settings())
        vif.create_input_files(Path.cwd())
        if hasattr(self, "_file_transfers"):
            self._file_transfers.transfer()

    def _structure(self):
        # avoid overlapping structure for e.g., phonon forces.
        if self.args.prev_dir and self.task.fine_to_inherit_structure_from_prev:
            return Structure.from_file(self.args.prev_dir / defaults.contcar)
        return Structure.from_file(self.args.poscar)

    def _overridden_incar_settings(self):
        result = deepcopy(defaults.user_incar_settings)
        result.update(self._prior_info.incar)

        if self.args.user_incar_settings:
            args = list2dict(self.args.user_incar_settings, all_incar_flags)
            result.update(args)

        return result

    def _overridden_potcar(self):
        result = {}
        if self.args.overridden_potcar:
            result.update(potcar_str2dict(self.args.overridden_potcar))
        return result

    def _option_kwargs(self):
        result = deepcopy(defaults.options)
        if self._prior_info:
            result.update(self._prior_info.input_options_kwargs)

        if self.args.prev_dir:
            pi = prior_info_from_calc_dir(prev_dir_path=self.args.prev_dir,
                                          vasprun=self.args.vasprun,
                                          outcar=self.args.outcar)
            result.update(pi.input_options_kwargs)

            self._file_transfers = FileTransfers(self._file_transfer(),
                                                 path=self.args.prev_dir)

        if self.args.options:
            args = list2dict(self.args.options, assignable_option_set)
            result.update(args)
        if self.args.uniform_kpt_mode:
            result["kpt_mode"] = KpointsMode.uniform

        return result

    def _file_transfer(self):
        if not self.args.file_transfer_type:
            return {}
        else:
            result = {}
            for filename, transfer in zip(self.args.file_transfer_type[0::2],
                                          self.args.file_transfer_type[1::2]):
                result[filename] = transfer
            return result


def plot_band(args: Namespace):
    band_plot_info_from_vasp = BandPlotInfoFromVasp(
        vasprun=Vasprun(args.vasprun), kpoints_filename=args.kpoints_filename)
    plot_info = band_plot_info_from_vasp.make_band_plot_info()
    plot_info.to_json_file()
    plotter = BandPlotter(plot_info, energy_range=args.y_range)
    plotter.construct_plot()
    plotter.plt.savefig(args.filename, format="pdf")


def plot_dos(args: Namespace):
    vasprun = Vasprun(args.vasprun)
    outcar = Outcar(args.outcar)
    band_edge = VaspBandEdgeProperties(vasprun, outcar)

    if band_edge.band_gap:
        vertical_lines = [band_edge.vbm_info.energy, band_edge.cbm_info.energy]
    else:
        vertical_lines = [vasprun.efermi]

    if args.base_energy is None:
        base = vertical_lines[0]
    else:
        base = args.base_energy

    dos_data_from_vasp = \
        DosDataFromVasp(vasprun, vertical_lines, base, args.crop_first_value)
    dos_data = dos_data_from_vasp.make_dos_data()

    ylim_set = None
    if args.y_max_ranges:
        if dos_data.spin:
            ylim_set = [[-y_max, y_max] for y_max in args.y_max_ranges]
        else:
            ylim_set = [[0, y_max] for y_max in args.y_max_ranges]

    structure = vasprun.final_structure
    grouped_atom_indices = args.type.grouped_atom_indices(structure, args.target)
    logger.info(f"Grouped atom indices: {grouped_atom_indices}")
    plot_data = dos_data.dos_plot_data(grouped_atom_indices,
                                       xlim=args.x_range,
                                       ylim_set=ylim_set)
    plot_data.to_json_file()
    plotter = DosPlotter(plot_data, args.legend)
    plotter.construct_plot()
    plotter.plt.savefig(args.filename, format="pdf")


def plot_absorption(args: Namespace):
    diele_func_data = make_diele_func(Vasprun(args.vasprun),
                                      Outcar(args.outcar),
                                      use_vasp_real=not args.calc_kk,
                                      ita=args.ita)
    diele_func_data.to_json_file()
    plotter = AbsorptionCoeffMplPlotter(diele_func_data, yranges=args.y_ranges)
    plotter.construct_plot()
    plotter.plt.savefig(args.filename, format="pdf")


def calc_effective_mass(args: Namespace):
    vasprun, outcar = Vasprun(args.vasprun), Outcar(args.outcar)
    band_edge_prop = VaspBandEdgeProperties(vasprun, outcar)
    try:
        vbm, cbm = band_edge_prop.vbm_cbm
    except TypeError:
        logger.warning("Band gap does not exist, so not suited for effective"
                       "mass calculation.")
        return
    effective_mass = make_effective_mass(vasprun,
                                         args.temperature,
                                         args.concentrations,
                                         vbm, cbm)
    print(effective_mass)


def band_edge_properties(args: Namespace):
    vasprun = Vasprun(args.vasprun)
    outcar = Outcar(args.outcar)
    print(VaspBandEdgeProperties(vasprun, outcar))



# -*- coding: utf-8 -*-

#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import re
from copy import deepcopy
from typing import List
import numpy as np

from pymatgen.electronic_structure.plotter import BSPlotter
from pymatgen.io.vasp import Vasprun
from vise.analyzer.plot_band import BandPlotInfo, BandInfo, XTicks, BandEdge
from vise.analyzer.plot_brillouin_zone import BZPlotInfo
from vise.util.string import latexify


def greek_to_unicode(label: str) -> str:
    d = {"GAMMA": "Γ", "SIGMA": "Σ", "DELTA": "Δ"}
    for k, v in d.items():
        label = label.replace(k, v)
    return label


def italic_to_roman(label: str) -> str:
    return re.sub(r"([A-Z])_([0-9])", r"{\\rm \1}_\2", label)


class BandPlotInfoFromVasp:
    def __init__(self,
                 vasprun: Vasprun,
                 kpoints_filename: str,
                 vasprun2: Vasprun = None,
                 energy_window: List[float] = None):
        self.vasprun = vasprun
        self.kpoints_filename = kpoints_filename
        self.vasprun2 = vasprun2
        self.energy_window = energy_window
        self.bs = self.vasprun.get_band_structure(self.kpoints_filename,
                                                  line_mode=True)

    def make_band_plot_info(self):
        bs_plotter = BSPlotter(self.bs)
        plot_data = bs_plotter.bs_plot_data(zero_to_efermi=False)
        distances = [list(d) for d in plot_data["distances"]]
        self._composition = self.vasprun.final_structure.composition

        band_info = [BandInfo(band_energies=self._remove_spin_key(plot_data),
                              band_edge=self._band_edge(self.bs, plot_data),
                              fermi_level=self.bs.efermi)]

        if self.vasprun2:
            bs2 = self.vasprun2.get_band_structure(self.kpoints_filename,
                                                   line_mode=True)
            plot_data2 = BSPlotter(bs2).bs_plot_data(zero_to_efermi=False)
            band_info.append(
                BandInfo(band_energies=self._remove_spin_key(plot_data2),
                         band_edge=self._band_edge(bs2, plot_data2),
                         fermi_level=self.bs.efermi))

        x = bs_plotter.get_ticks_old()
        x_ticks = XTicks(_sanitize_labels(x["label"]), x["distance"])

        return BandPlotInfo(band_info_set=band_info,
                            distances_by_branch=distances,
                            x_ticks=x_ticks,
                            title=self._title)

    def make_bz_plot_info(self):
        rec_lat = self.vasprun.final_structure.lattice.reciprocal_lattice
        faces = [[[float(k) for k in j] for j in i] for i in rec_lat.get_wigner_seitz_cell()]
        labels = {}

        concat = False
        band_paths = []
        init_point = None
        for kpoint in self.bs.kpoints:
            if kpoint.label:
                c_coords = list(kpoint.cart_coords)
                f_coords = list(kpoint.frac_coords)
                label = greek_to_unicode(kpoint.label)
                labels[label] = {"cart": c_coords, "frac": f_coords}
                if concat is False and init_point:
                    band_paths.append([init_point, c_coords])
                init_point = c_coords
                concat = True
            else:
                concat = False

        return BZPlotInfo(faces, labels, band_paths, rec_lat.matrix.tolist())

    def _remove_spin_key(self, plot_data) -> List[List[List[List[float]]]]:
        """
        Pymatgen at 2020.11.11
         energy: A dict storing bands for spin up and spin down data
            {Spin:[np.array(nb_bands,kpoints),...]} as a list of discontinuous kpath
            of energies. The energy of multiple continuous branches are stored together.

        -> [branch][spin][band][k-point]
        """
        num_spin = len(plot_data["energy"])
        num_branch = len(plot_data["energy"]["1"])

        result = [[[] for _ in range(num_spin)] for __ in range(num_branch)]
        for spin_idx, (_, branch_energies) in enumerate(
                sorted(plot_data["energy"].items(),
                       key=lambda item: item[0],
                       reverse=True)):
            for branch_idx, branch_energy in enumerate(branch_energies):
                if self.energy_window:
                    removed_idxs = []
                    for i in range(len(branch_energy)):
                        _max = np.max(branch_energy[i, :])
                        _min = np.min(branch_energy[i, :])
                        if not self.in_energy(_max, _min):
                            removed_idxs.append(i)
                    x = np.delete(branch_energy, removed_idxs, axis=0).tolist()
                else:
                    x = branch_energy.tolist()

                result[branch_idx][spin_idx] = deepcopy(x)

        return result

    def in_energy(self, _max, _min):
        return _max >= self.energy_window[0] and _min <= self.energy_window[1]

    def _band_edge(self, bs, plot_data):
        if bs.is_metal():
            return None
        else:
            return BandEdge(
                vbm=plot_data["vbm"][0][1],
                cbm=plot_data["cbm"][0][1],
                vbm_distances=[i[0] for i in plot_data["vbm"]],
                cbm_distances=[i[0] for i in plot_data["cbm"]])

    @property
    def _title(self):
        return latexify(self._composition.reduced_formula)


def _sanitize_label(label):
    return italic_to_roman(greek_to_unicode(label))


def _sanitize_labels(labels):
    return [_sanitize_label(label) for label in labels]



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from dataclasses import dataclass
from math import sqrt, pi
from typing import List

import numpy as np
from monty.json import MSONable
from tqdm import tqdm
from vise.util.mix_in import ToJsonFileMixIn
from scipy.constants import physical_constants as pc

eV_to_inv_cm = pc["electron volt-inverse meter relationship"][0] / 100


def diele_func_to_coeff(freq, real, imag):
    return (2 * sqrt(2) * pi * sqrt(sqrt(real ** 2 + imag ** 2) - real)
            * freq * eV_to_inv_cm)


@dataclass
class DieleFuncData(MSONable, ToJsonFileMixIn):
    energies: List[float]  # in eV
    diele_func_real: List[List[float]]  # [xx, yy, zz, xy, yz, xz]
    diele_func_imag: List[List[float]]  # [xx, yy, zz, xy, yz, xz]
    band_gap: float  # in eV

    @property
    def ave_absorption_coeff(self):
        reals = [sum(self.diele_func_real[i][:3]) / 3
                 for i in range(len(self.energies))]
        imags = [sum(self.diele_func_imag[i][:3]) / 3
                 for i in range(len(self.energies))]
        return [diele_func_to_coeff(freq, real, imag)
                for freq, real, imag in zip(self.energies, reals, imags)]

    def target_coeff_min_e(self, target_coeff: float = 10**4):
        for e, coeff in zip(self.energies, self.ave_absorption_coeff):
            if coeff > target_coeff:
                return e
        return None


def make_shifted_diele_func(diele_func_data: DieleFuncData,
                            original_band_gap: float,
                            shift: float) -> DieleFuncData:
    imag = imag_shift(diele_func_data.diele_func_imag,
                      diele_func_data.energies,
                      original_band_gap + shift, shift)
    real = kramers_kronig_trans(imag, diele_func_data.energies)
    return DieleFuncData(diele_func_data.energies,
                         real.tolist(),
                         imag.tolist(),
                         original_band_gap + shift)


def imag_shift(diele_func_imag: List[List[float]],
               energies: List[float],
               band_gap: float,
               shift: float) -> np.ndarray:
    energies = np.array(energies)
    assert shift > 0
    result = []
    for energy_grid in energies:
        old_e = energy_grid - shift
        right_idx = np.argwhere(energies > old_e)[0][0]
        left_e, right_e = energies[right_idx - 1], energies[right_idx]
        # linear interpolation
        left_ratio = (right_e - old_e) / (right_e - left_e)

        inner_result = []
        for imag_idx in range(6):
            if energy_grid < band_gap:
                inner_result.append(0.0)
            else:
                old_diele = \
                    diele_func_imag[right_idx - 1][imag_idx] * left_ratio + \
                    diele_func_imag[right_idx][imag_idx] * (1 - left_ratio)
                inner_result.append(
                    old_diele * (energy_grid - shift) / energy_grid)

        result.append(inner_result)

    return np.array(result)


def kramers_kronig_trans(diele_func_imag: np.array,
                         energies: List[float],
                         ita: float = 0.01) -> np.ndarray:
    mesh = energies[1] - energies[0]
    result = []
    ee2ss = [[e ** 2 - energy_grid ** 2 for e in energies]
             for energy_grid in energies]
    for imag_idx in tqdm(range(6)):
        imags = diele_func_imag[:, imag_idx]
        if imag_idx == 0 or \
                (imag_idx > 0
                 and np.allclose(
                            imags, diele_func_imag[:, imag_idx - 1]) is False):
            if np.count_nonzero(imags) == 0:
                inner_result = [0.0] * len(energies)
            else:
                inner_result = []
                for ee2s in ee2ss:
                    integrals = [e * imag * ee2 / (ee2 ** 2 + ita ** 2)
                                 for e, ee2, imag in zip(energies, ee2s, imags)]
                    integral = sum(integrals) * mesh * 2 / pi
                    if imag_idx < 3:
                        integral += 1
                    inner_result.append(integral)

        result.append(inner_result)

    return np.array(result).T


# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import re
from copy import deepcopy
from pathlib import Path
from typing import List

import yaml
from vise.util.logger import get_logger
logger = get_logger(__name__)


class UserSettings:
    def __init__(self, yaml_filename: str):
        self._cwd = Path.cwd()
        self.yaml_filename = yaml_filename
        self.yaml_files_from_root_dir = self._make_yaml_file_list()

    def _make_yaml_file_list(self) -> List[Path]:
        result = []

        dirname = self._cwd
        while True:
            filenames = [self.yaml_filename, "." + self.yaml_filename]
            file_paths = [dirname / filename for filename in filenames]
            for file_path in file_paths:
                if file_path.exists():
                    result.append(file_path)

            if dirname == Path("/"):
                break
            else:
                dirname = dirname.parent

        return list(reversed(result))

    @property
    def user_settings(self) -> dict:
        result = {}

        for file_path in self.yaml_files_from_root_dir:
            logger.info(f"Setting file: {file_path} is parsed...")
            try:
                with open(str(file_path), "r") as fin:
                    settings = yaml.load(fin, Loader=yaml.SafeLoader)
                    result.update(self._add_absolute_path(settings, file_path))
            except AttributeError:
                pass

        return result

    def _add_absolute_path(self, settings, file_path):
        result = deepcopy(settings)
        for key, value in settings.items():
            if self.is_path(value):
                result[key] = file_path.parent / value

        return result

    @staticmethod
    def is_path(key):
        return isinstance(key, str) and re.match(r'\S*/\S*', key)



# -*- coding: utf-8 -*-
#  Copyright (c) 2021. Distributed under the terms of the MIT License.
from dataclasses import dataclass
from typing import List, Tuple

import pytest
from vise.tests.helpers.assertion import assert_dataclass_almost_equal


@dataclass
class Data:
    x: float


@dataclass
class TestData:
    a: float = None
    b: List[float] = None
    c: List[Tuple[float]] = None
    d: Data = None


def test_assert_dataclass_almost_equal_float():
    with pytest.raises(AssertionError):
        assert_dataclass_almost_equal(TestData(1.0), 1.0)

    assert_dataclass_almost_equal(TestData(1.0), TestData(1.0))

    with pytest.raises(AssertionError):
        assert_dataclass_almost_equal(TestData(1.0), TestData(2.0))

    with pytest.raises(AssertionError):
        assert_dataclass_almost_equal(TestData(1.0), TestData(1.01), digit=2)

    assert_dataclass_almost_equal(TestData(1.0), TestData(1.01), digit=1)


def test_assert_dataclass_almost_equal_iterable():
    assert_dataclass_almost_equal(TestData(b=[1.0]), TestData(b=[1.0]))

    with pytest.raises(AssertionError):
        assert_dataclass_almost_equal(TestData(b=[1.0]), TestData(b=[2.0]))


def test_assert_dataclass_almost_equal_nested_iterable():
    assert_dataclass_almost_equal(TestData(c=[(1.0,)]), TestData(c=[(1.0,)]))
    with pytest.raises(AssertionError):
        assert_dataclass_almost_equal(TestData(c=[(1.0,)]),
                                      TestData(c=[(2.0,)]), digit=2)


def test_assert_dataclass_almost_equal_dataclass_attribute():
    assert_dataclass_almost_equal(TestData(d=Data(1.0)), TestData(d=Data(1.0)))
    with pytest.raises(AssertionError):
        assert_dataclass_almost_equal(TestData(d=Data(1.0)), TestData(d=Data(1.01)), digit=2)

    assert_dataclass_almost_equal(TestData(d=Data(1.0)), TestData(d=Data(1.001)), digit=2)





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import itertools
from dataclasses import dataclass
from math import sqrt
from typing import List, Dict

from monty.json import MSONable

import plotly.graph_objects as go
from vise.analyzer.plot_band_dos import plotly_sanitize_label
from vise.util.plotly_util import sort_coords, make_triangles
import numpy as np


@dataclass
class BZPlotInfo(MSONable):
    faces: List[List[List[float]]]
    # {"X": {"cart": [0.5, 0, 0], "frac": [0.7514, 0, 0]"}}
    labels: Dict[str, Dict[str, List[float]]]
    band_paths: List[List[List[float]]] = None
    rec_lat_vec: List[List[float]] = None


def pairwise(iterable):
    """s -> (s0,s1), (s1,s2), (s2, s3), ..."""
    a, b = itertools.tee(iterable)
    next(b)
    return itertools.zip_longest(a, b, fillvalue=iterable[0])


class BZPlotlyPlotter:
    def __init__(self, bz_plot_info: BZPlotInfo):
        self._bz_plot_info = bz_plot_info

    def create_figure(self):
        data = []
        for face_vertices in self._bz_plot_info.faces:
            vertex_coords = sort_coords(np.array(face_vertices))
            # data.append(go.Mesh3d(**make_triangles(vertex_coords),
            #                       alphahull=-1, opacity=0.1, hoverinfo='skip',
            #                       color="blue"))
            for s, t in pairwise(vertex_coords):
                data.append(go.Scatter3d(x=[s[0], t[0]],
                                         y=[s[1], t[1]],
                                         z=[s[2], t[2]], mode="lines",
                                         hoverinfo="none",
                                         marker_color="black"))

        for label, coords in self._bz_plot_info.labels.items():
            c_coords, f_coords = coords["cart"], coords["frac"]
            data.append(go.Scatter3d(x=[c_coords[0]], y=[c_coords[1]], z=[c_coords[2]],
                                     text=[plotly_sanitize_label(label)],
                                     mode="markers+text",
                                     marker=dict(color="orange", size=6),
                                     meta=f_coords,
                                     hovertemplate="(%{meta[0]:.2f} %{meta[1]:.2f} %{meta[2]:.2f})",
                                     textposition="middle center",
                                     textfont=dict(size=32)))

        for i, l in self._bz_plot_info.band_paths:
            data.append(go.Scatter3d(x=[i[0], l[0]],
                                     y=[i[1], l[1]],
                                     z=[i[2], l[2]], mode="lines",
                                     opacity=0.5,
                                     hoverinfo="none",
                                     line_width=8,
                                     marker_color="purple"))

        _max = np.amax(np.array(sum(self._bz_plot_info.faces, [])))
        c_max = _max * 1.3
        c_cone = _max * 0.3

        data.append(go.Cone(x=[c_max], y=[0], z=[0], u=[c_cone], v=[0], w=[0],
                            colorscale=[[0, 'rgb(200,200,200)'],
                                        [1, 'rgb(200,200,200)']],
                            showscale=False, hovertemplate="<b>x</b>"))
        data.append(go.Cone(x=[0], y=[c_max], z=[0], u=[0], v=[c_cone], w=[0],
                            colorscale=[[0, 'rgb(200,200,200)'],
                                        [1, 'rgb(200,200,200)']],
                            showscale=False, hovertemplate="<b>y</b>"))
        data.append(go.Cone(x=[0], y=[0], z=[c_max], u=[0], v=[0], w=[c_cone],
                            colorscale=[[0, 'rgb(200,200,200)'],
                                        [1, 'rgb(200,200,200)']],
                            showscale=False, hovertemplate="<b>z</b>"))
        data.append(go.Scatter3d(x=[0, c_max], y=[0, 0], z=[0, 0], mode="lines",
                                 hoverinfo="none",
                                 marker_color="black"))
        data.append(go.Scatter3d(x=[0, 0], y=[0, c_max], z=[0, 0], mode="lines",
                                 hoverinfo="none",
                                 marker_color="black"))
        data.append(go.Scatter3d(x=[0, 0], y=[0, 0], z=[0, c_max], mode="lines",
                                 hoverinfo="none",
                                 marker_color="black"))

        for i, direct in zip(self._bz_plot_info.rec_lat_vec, ["kx", "ky", "kz"]):
            nn = sqrt(i[0] ** 2 + i[1] ** 2 + i[2] ** 2)
            kx, ky, kz = np.array(i) * c_max / nn * 1.15
            norm_kx, norm_ky, norm_kz = np.array(i) * c_cone / nn
            data.append(go.Cone(x=[kx], y=[ky], z=[kz],
                                u=[norm_kx], v=[norm_ky], w=[norm_kz],
                                colorscale=[[0, 'rgb(100,100,100)'],
                                            [1, 'rgb(100,100,100)']],
                                showscale=False, hovertemplate=f"<b>{direct}</b>"))
            data.append(go.Scatter3d(x=[0, kx], y=[0, ky], z=[0, kz],
                                     mode="lines",
                                     hoverinfo="none",
                                     marker_color="black"))

        range_max = c_max * 1.4
        fig = go.Figure(data=data)

        fig.update_layout(
            title_font_size=30,
            font_size=24,
            width=900, height=900,
            showlegend=False,
            scene=dict(xaxis=dict(showspikes=False, range=[-range_max, range_max],
                                  showticklabels=False, visible=False),
                       yaxis=dict(showspikes=False, range=[-range_max, range_max],
                                  showticklabels=False, visible=False),
                       zaxis=dict(showspikes=False, range=[-range_max, range_max],
                                  showticklabels=False, visible=False))
        )

        return fig





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import dash
from pymatgen.core import Structure, Lattice
from vise.analyzer.dash_components.structure_component import StructureComponent
import dash_html_components as html

import crystal_toolkit.components as ctc

app = dash.Dash()
ctc.register_app(app)

structure = Structure(lattice=Lattice.cubic(10), species=["H"], coords=[[0.0]*3])

component = StructureComponent(structure)
my_layout = html.Div(
    [
        component.layout(),
    ]
)

# app.layout = ctc.crystal_toolkit_layout(my_layout)
ctc.register_crystal_toolkit(app=app, layout=my_layout, cache=None)


if __name__ == "__main__":
    app.run_server(debug=True, port=8051)




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from math import log10
from typing import List

from matplotlib import pyplot as plt
from plotly import graph_objects as go
from vise.analyzer.dielectric_function import DieleFuncData
from vise.analyzer.dielectric_function_data.exp_dielectric_func import \
    ExpDieleFunc
from vise.util.matplotlib import float_to_int_formatter


class AbsorptionCoeffPlotter:

    def __init__(self,
                 diele_func_data: DieleFuncData,
                 energy_range: List[float] = None,
                 coeff_power_range: List[float] = None,
                 materials: List[str] = None,
                 yranges: List[float] = None):

        self.energies = diele_func_data.energies
        self.absorption_coeff = diele_func_data.ave_absorption_coeff
        self.band_gap = diele_func_data.band_gap
        self.energy_range = energy_range or [0, 10]
        self.coeff_power_range = coeff_power_range
        self.materials = materials

        self._xaxis_title = "Energy (eV)"

        if yranges:
            self.ymin, self.ymax = yranges
        else:
            self.ymin, self.ymax = 10**3, 10**7

        self.plt = plt


class AbsorptionCoeffPlotlyPlotter(AbsorptionCoeffPlotter):

    _yaxis_title = "Absorption coefficient. (cm <sup>-1</sup>)"

    def create_figure(self):
        fig = go.Figure()
        fig.update_layout(
            xaxis_title=self._xaxis_title,
            yaxis_title=self._yaxis_title,
            font_size=25,
            width=800, height=700)

        fig.add_trace(go.Scatter(x=self.energies,
                                 y=self.absorption_coeff,
                                 line=dict(width=2.5),
                                 name="Average"))
        fig.add_trace(go.Scatter(x=[self.band_gap, self.band_gap],
                                 y=[self.ymin, self.ymax],
                                 line=dict(width=2, dash="dash"),
                                 line_color="black",
                                 name="Band gap"))

        if self.materials:
            for material in self.materials:
                exp = ExpDieleFunc(material)
                energies = exp.energies
                fig.add_trace(go.Scatter(x=[exp.band_gap, exp.band_gap],
                                         y=[self.ymin, self.ymax],
                                         line=dict(width=1, dash="dash"),
                                         showlegend=False,
                                         line_color="black",
                                         name=f"{material} band gap"))

                fig.add_trace(go.Scatter(x=energies, y=exp.absorption_coeff,
                                         line=dict(width=1, dash="dashdot"),
                                         name=material))

        fig.update_xaxes(range=self.energy_range, tickfont_size=20)
        fig.update_yaxes(type="log",
                         range=[log10(self.ymin), log10(self.ymax)],
                         tickfont_size=20,
                         showexponent="all", exponentformat='power'
                         )
        return fig


class AbsorptionCoeffMplPlotter(AbsorptionCoeffPlotter):
    _yaxis_title = "Absorption coefficient. (cm-1)"

    def construct_plot(self):
        self._add_coeffs()
        self._add_band_gap()
        if self.materials:
            self._add_materials()
        # self._set_figure_legend()
        self._set_x_range()
        self._set_y_range()
        self._set_labels()
        self._set_formatter()
        self.plt.tight_layout()

    def _add_coeffs(self):
        self.plt.semilogy(self.energies, self.absorption_coeff)

    def _add_band_gap(self):
        self.plt.axvline(x=self.band_gap, linestyle="dashed", color="black",
                         linewidth=1)

    def _add_materials(self):
        for material in self.materials:
            exp = ExpDieleFunc(material)
            energies = exp.energies
            self.plt.axvline(x=exp.band_gap,
                             linestyle="dashed",
                             color="black",
                             linewidth=1)

            self.plt.semilogy(energies,
                              list(exp.absorption_coeff),
                              label=material,
                              linestyle="dashdot",
                              )

    def _set_figure_legend(self):
        self.plt.legend(loc="lower right")

    def _set_x_range(self):
        self.plt.xlim(0, 10)

    def _set_y_range(self):
        self.plt.gca().set_ylim(ymin=self.ymin, ymax=self.ymax)

    def _set_labels(self):
        self.plt.xlabel(self._xaxis_title)
        self.plt.ylabel(self._yaxis_title)

    def _set_formatter(self):
        axis = self.plt.gca()
        axis.xaxis.set_major_formatter(float_to_int_formatter)


# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from dataclasses import dataclass

from monty.json import MSONable
from monty.serialization import loadfn
from vise.util.mix_in import ToJsonFileMixIn, ToYamlFileMixIn


@dataclass
class TestTest(MSONable, ToJsonFileMixIn, ToYamlFileMixIn):
    a: str = "aaa"

    def to_yaml(self):
        return f"a: {self.a}\n"

    @classmethod
    def from_yaml(cls, filename):
        d = loadfn(filename)
        return cls(**d)


def test_to_json_file_mix_in(tmpdir):
    actual = TestTest()
    tmpdir.chdir()
    actual.to_json_file()
    actual = loadfn("test_test.json")
    assert actual == TestTest()


def test_to_yaml_file_mix_in(tmpdir):
    actual = TestTest()
    tmpdir.chdir()
    print(tmpdir)
    actual.to_yaml_file()
    actual = TestTest.from_yaml("test_test.yaml")
    assert actual == TestTest()




# -*- coding: utf-8 -*-

from itertools import groupby
import operator

from pymatgen.core import Structure

from vise.util.logger import get_logger


logger = get_logger(__name__)


def create_symbol_list(structure: Structure):
    species = [str(s) for s in structure.species]
    # unique_justseen https://docs.python.org/ja/3/library/itertools.html
    # ["H", "H", "O", "O", "H"] -> ['H', 'O', 'H']
    return list(map(next, map(operator.itemgetter(1), groupby(species, None))))


